#    Copyright 2015-2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import time

import msgpack
import six
import testscenarios

from oslo_config import cfg

import oslo_messaging
from oslo_messaging._drivers.zmq_driver.proxy.central \
    import zmq_publisher_proxy
from oslo_messaging._drivers.zmq_driver.proxy import zmq_proxy
from oslo_messaging._drivers.zmq_driver import zmq_address
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._drivers.zmq_driver import zmq_version
from oslo_messaging.tests.drivers.zmq import zmq_common

load_tests = testscenarios.load_tests_apply_scenarios

zmq = zmq_async.import_zmq()

opt_group = cfg.OptGroup(name='zmq_proxy_opts',
                         title='ZeroMQ proxy options')
cfg.CONF.register_opts(zmq_proxy.zmq_proxy_opts, group=opt_group)


class TestPubSub(zmq_common.ZmqBaseTestCase):

    LISTENERS_COUNT = 3

    scenarios = [
        ('json', {'serialization': 'json',
                  'dumps': lambda obj: six.b(json.dumps(obj))}),
        ('msgpack', {'serialization': 'msgpack',
                     'dumps': msgpack.dumps})
    ]

    def setUp(self):
        super(TestPubSub, self).setUp()

        kwargs = {'use_pub_sub': True,
                  'rpc_zmq_serialization': self.serialization}
        self.config(group='oslo_messaging_zmq', **kwargs)

        self.config(host="127.0.0.1", group="zmq_proxy_opts")
        self.config(publisher_port=0, group="zmq_proxy_opts")

        self.publisher = zmq_publisher_proxy.PublisherProxy(
            self.conf, self.driver.matchmaker)
        self.driver.matchmaker.register_publisher(
            (self.publisher.host, ''),
            expire=self.conf.oslo_messaging_zmq.zmq_target_expire)

        self.listeners = []
        for _ in range(self.LISTENERS_COUNT):
            self.listeners.append(zmq_common.TestServerListener(self.driver))

    def tearDown(self):
        super(TestPubSub, self).tearDown()
        self.publisher.cleanup()
        for listener in self.listeners:
            listener.stop()

    def _send_request(self, target):
        #  Needed only in test env to give listener a chance to connect
        #  before request fires
        time.sleep(1)
        context = {}
        message = {'method': 'hello-world'}

        self.publisher.send_request(
            [b"reply_id",
             b'',
             six.b(zmq_version.MESSAGE_VERSION),
             six.b(str(zmq_names.CAST_FANOUT_TYPE)),
             zmq_address.target_to_subscribe_filter(target),
             b"message_id",
             self.dumps([context, message])]
        )

    def _check_listener(self, listener):
        listener._received.wait(timeout=5)
        self.assertTrue(listener._received.isSet())
        method = listener.message.message[u'method']
        self.assertEqual(u'hello-world', method)

    def _check_listener_negative(self, listener):
        listener._received.wait(timeout=1)
        self.assertFalse(listener._received.isSet())

    def test_single_listener(self):
        target = oslo_messaging.Target(topic='testtopic', fanout=True)
        self.listener.listen(target)

        self._send_request(target)

        self._check_listener(self.listener)

    def test_all_listeners(self):
        target = oslo_messaging.Target(topic='testtopic', fanout=True)

        for listener in self.listeners:
            listener.listen(target)

        self._send_request(target)

        for listener in self.listeners:
            self._check_listener(listener)

    def test_filtered(self):
        target = oslo_messaging.Target(topic='testtopic', fanout=True)
        target_wrong = oslo_messaging.Target(topic='wrong', fanout=True)

        self.listeners[0].listen(target)
        self.listeners[1].listen(target)
        self.listeners[2].listen(target_wrong)

        self._send_request(target)

        self._check_listener(self.listeners[0])
        self._check_listener(self.listeners[1])
        self._check_listener_negative(self.listeners[2])

    def test_topic_part_matching(self):
        target = oslo_messaging.Target(topic='testtopic', server='server')
        target_part = oslo_messaging.Target(topic='testtopic', fanout=True)

        self.listeners[0].listen(target)
        self.listeners[1].listen(target)

        self._send_request(target_part)

        self._check_listener(self.listeners[0])
        self._check_listener(self.listeners[1])



#    Copyright 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools

import oslo_messaging
from oslo_messaging._drivers import impl_zmq
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_socket
from oslo_messaging.tests.drivers.zmq import zmq_common
from oslo_messaging.tests import utils as test_utils


zmq = zmq_async.import_zmq()


class ZmqTestPortsRange(zmq_common.ZmqBaseTestCase):

    @testtools.skipIf(zmq is None, "zmq not available")
    def setUp(self):
        super(ZmqTestPortsRange, self).setUp()

        # Set config values
        kwargs = {'rpc_zmq_min_port': 5555,
                  'rpc_zmq_max_port': 5560}
        self.config(group='oslo_messaging_zmq', **kwargs)

    def test_ports_range(self):
        listeners = []

        for i in range(10):
            try:
                target = oslo_messaging.Target(topic='testtopic_' + str(i))
                new_listener = self.driver.listen(target, None, None)
                listeners.append(new_listener)
            except zmq_socket.ZmqPortBusy:
                pass

        self.assertLessEqual(len(listeners), 5)

        for l in listeners:
            l.cleanup()


class TestConfZmqDriverLoad(test_utils.BaseTestCase):

    @testtools.skipIf(zmq is None, "zmq not available")
    def setUp(self):
        super(TestConfZmqDriverLoad, self).setUp()
        self.messaging_conf.transport_driver = 'zmq'

    def test_driver_load(self):
        transport = oslo_messaging.get_transport(self.conf)
        self.assertIsInstance(transport._driver, impl_zmq.ZmqDriver)


class TestZmqBasics(zmq_common.ZmqBaseTestCase):

    @testtools.skipIf(zmq is None, "zmq not available")
    def setUp(self):
        super(TestZmqBasics, self).setUp()
        self.target = oslo_messaging.Target(topic='topic')
        self.ctxt = {'key': 'value'}
        self.message = {'method': 'qwerty', 'args': {'int': 1, 'bool': True}}

    def test_send_call_without_method_failure(self):
        self.message.pop('method')
        self.listener.listen(self.target)
        self.assertRaises(KeyError, self.driver.send,
                          self.target, self.ctxt, self.message,
                          wait_for_reply=True, timeout=10)

    def _check_listener_received(self):
        self.assertTrue(self.listener._received.isSet())
        self.assertEqual(self.ctxt, self.listener.message.ctxt)
        self.assertEqual(self.message, self.listener.message.message)

    def test_send_call_success(self):
        self.listener.listen(self.target)
        result = self.driver.send(self.target, self.ctxt, self.message,
                                  wait_for_reply=True, timeout=10)
        self.assertTrue(result)
        self._check_listener_received()

    def test_send_call_direct_success(self):
        self.target.server = 'server'
        self.listener.listen(self.target)
        result = self.driver.send(self.target, self.ctxt, self.message,
                                  wait_for_reply=True, timeout=10)
        self.assertTrue(result)
        self._check_listener_received()

    def test_send_cast_direct_success(self):
        self.target.server = 'server'
        self.listener.listen(self.target)
        result = self.driver.send(self.target, self.ctxt, self.message,
                                  wait_for_reply=False)
        self.listener._received.wait(5)
        self.assertIsNone(result)
        self._check_listener_received()

    def test_send_fanout_success(self):
        self.target.fanout = True
        self.listener.listen(self.target)
        result = self.driver.send(self.target, self.ctxt, self.message,
                                  wait_for_reply=False)
        self.listener._received.wait(5)
        self.assertIsNone(result)
        self._check_listener_received()

    def test_send_notify_success(self):
        self.listener.listen_notifications([(self.target, 'info')])
        self.target.topic += '.info'
        result = self.driver.send_notification(self.target, self.ctxt,
                                               self.message, '3.0')
        self.listener._received.wait(5)
        self.assertIsNone(result)
        self._check_listener_received()



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import contextlib
import logging

import six
import tenacity

from oslo_messaging._drivers.zmq_driver.matchmaker import zmq_matchmaker_base
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._i18n import _LW

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


def _drop_message_warn(request):
    LOG.warning(_LW("Matchmaker contains no records for specified "
                    "target %(target)s. Dropping message %(msg_id)s."),
                {"target": request.target,
                 "msg_id": request.message_id})


def target_not_found_warn(func):
    def _target_not_found_warn(self, request, *args, **kwargs):
        try:
            return func(self, request, *args, **kwargs)
        except (zmq_matchmaker_base.MatchmakerUnavailable,
                tenacity.RetryError):
            _drop_message_warn(request)
    return _target_not_found_warn


def target_not_found_timeout(func):
    def _target_not_found_timeout(self, request, *args, **kwargs):
        try:
            return func(self, request, *args, **kwargs)
        except (zmq_matchmaker_base.MatchmakerUnavailable,
                tenacity.RetryError):
            _drop_message_warn(request)
            self.publisher._raise_timeout(request)
    return _target_not_found_timeout


@six.add_metaclass(abc.ABCMeta)
class PublisherManagerBase(object):

    """Abstract publisher manager class

    Publisher knows how to establish connection, how to send message,
    and how to receive reply. PublisherManager coordinates all these steps
    regarding retrying logic in AckManager implementations. May also have an
    additional thread pool for scheduling background tasks.
    """

    def __init__(self, publisher, with_pool=False):
        self.publisher = publisher
        self.conf = publisher.conf
        self.sender = publisher.sender
        self.receiver = publisher.receiver
        if with_pool:
            self.pool = zmq_async.get_pool(
                size=self.conf.oslo_messaging_zmq.rpc_thread_pool_size
            )
        else:
            self.pool = None

    @abc.abstractmethod
    def send_call(self, request):
        """Send call request

        :param request: request object
        :type request: zmq_request.CallRequest
        """

    @abc.abstractmethod
    def send_cast(self, request):
        """Send cast request

        :param request: request object
        :type request: zmq_request.CastRequest
        """

    @abc.abstractmethod
    def send_fanout(self, request):
        """Send fanout request

        :param request: request object
        :type request: zmq_request.FanoutRequest
        """

    @abc.abstractmethod
    def send_notify(self, request):
        """Send notification request

        :param request: request object
        :type request: zmq_request.NotificationRequest
        """

    def cleanup(self):
        if self.pool:
            self.pool.shutdown(wait=True)
        self.publisher.cleanup()


class PublisherManagerDynamic(PublisherManagerBase):

    @target_not_found_timeout
    def send_call(self, request):
        with contextlib.closing(self.publisher.acquire_connection(request)) \
                as socket:
            self.publisher.send_request(socket, request)
            reply = self.publisher.receive_reply(socket, request)
            return reply

    @target_not_found_warn
    def _send(self, request):
        with contextlib.closing(self.publisher.acquire_connection(request)) \
                as socket:
            self.publisher.send_request(socket, request)

    send_cast = _send
    send_fanout = _send
    send_notify = _send


class PublisherManagerDynamicAsyncMultisend(PublisherManagerDynamic):

    def __init__(self, publisher):
        super(PublisherManagerDynamicAsyncMultisend, self).__init__(
            publisher, with_pool=True
        )

    def _send_async(self, request):
        self.pool.submit(self._send, request)

    send_fanout = _send_async
    send_notify = _send_async


class PublisherManagerStatic(PublisherManagerBase):

    @target_not_found_timeout
    def send_call(self, request):
        socket = self.publisher.acquire_connection(request)
        self.publisher.send_request(socket, request)
        reply = self.publisher.receive_reply(socket, request)
        return reply

    @target_not_found_warn
    def _send(self, request):
        socket = self.publisher.acquire_connection(request)
        self.publisher.send_request(socket, request)

    send_cast = _send
    send_fanout = _send
    send_notify = _send


class PublisherManagerStaticAsyncMultisend(PublisherManagerStatic):

    def __init__(self, publisher):
        super(PublisherManagerStaticAsyncMultisend, self).__init__(
            publisher, with_pool=True
        )

    def _send_async(self, request):
        self.pool.submit(self._send, request)

    send_fanout = _send_async
    send_notify = _send_async



#    Copyright 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import uuid

import six

from oslo_messaging._drivers import common as rpc_common
from oslo_messaging._drivers.zmq_driver import zmq_address
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._i18n import _LE
from oslo_messaging import exceptions
from oslo_serialization.serializer import json_serializer
from oslo_serialization.serializer import msgpack_serializer

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


class ZmqSocket(object):

    SERIALIZERS = {
        'json': json_serializer.JSONSerializer(),
        'msgpack': msgpack_serializer.MessagePackSerializer()
    }

    def __init__(self, conf, context, socket_type, immediate,
                 high_watermark=0, identity=None):
        self.conf = conf
        self.context = context
        self.socket_type = socket_type
        self.handle = context.socket(socket_type)
        self.handle.set_hwm(high_watermark)

        # Set linger period
        linger = -1
        if self.conf.oslo_messaging_zmq.zmq_linger >= 0:
            # Convert seconds to milliseconds
            linger = self.conf.oslo_messaging_zmq.zmq_linger * 1000
        self.handle.setsockopt(zmq.LINGER, linger)

        # Put messages to only connected queues
        self.handle.setsockopt(zmq.IMMEDIATE, 1 if immediate else 0)

        # Setup timeout on socket sending
        if hasattr(self.conf, 'rpc_response_timeout'):
            self.handle.setsockopt(zmq.SNDTIMEO,
                                   self.conf.rpc_response_timeout * 1000)

        # Configure TCP keep alive
        keepalive = self.conf.oslo_messaging_zmq.zmq_tcp_keepalive
        if keepalive < 0:
            keepalive = -1
        elif keepalive > 0:
            keepalive = 1
        self.handle.setsockopt(zmq.TCP_KEEPALIVE, keepalive)

        keepalive_idle = self.conf.oslo_messaging_zmq.zmq_tcp_keepalive_idle
        if keepalive_idle <= 0:
            keepalive_idle = -1
        self.handle.setsockopt(zmq.TCP_KEEPALIVE_IDLE, keepalive_idle)

        keepalive_cnt = self.conf.oslo_messaging_zmq.zmq_tcp_keepalive_cnt
        if keepalive_cnt <= 0:
            keepalive_cnt = -1
        self.handle.setsockopt(zmq.TCP_KEEPALIVE_CNT, keepalive_cnt)

        keepalive_intvl = self.conf.oslo_messaging_zmq.zmq_tcp_keepalive_intvl
        if keepalive_intvl <= 0:
            keepalive_intvl = -1
        self.handle.setsockopt(zmq.TCP_KEEPALIVE_INTVL, keepalive_intvl)

        self.handle.identity = \
            six.b(str(uuid.uuid4())) if identity is None else identity

        self.connections = set()

    def _get_serializer(self, serialization):
        serializer = self.SERIALIZERS.get(serialization, None)
        if serializer is None:
            raise NotImplementedError(
                "Serialization '{}' is not supported".format(serialization)
            )
        return serializer

    def type_name(self):
        return zmq_names.socket_type_str(self.socket_type)

    def connections_count(self):
        return len(self.connections)

    def connect(self, address):
        if address not in self.connections:
            self.handle.connect(address)
            self.connections.add(address)

    def setsockopt(self, *args, **kwargs):
        self.handle.setsockopt(*args, **kwargs)

    def setsockopt_string(self, *args, **kwargs):
        self.handle.setsockopt_string(*args, **kwargs)

    def getsockopt(self, *args, **kwargs):
        return self.handle.getsockopt(*args, **kwargs)

    def getsockopt_string(self, *args, **kwargs):
        return self.handle.getsockopt_string(*args, **kwargs)

    def send(self, *args, **kwargs):
        self.handle.send(*args, **kwargs)

    def send_string(self, u, *args, **kwargs):
        # NOTE(ozamiatin): Not using send_string until
        # eventlet zmq support this convenience method
        # in thread-safe manner
        encoding = kwargs.pop('encoding', 'utf-8')
        s = u.encode(encoding) if isinstance(u, six.text_type) else u
        self.handle.send(s, *args, **kwargs)

    def send_json(self, *args, **kwargs):
        self.handle.send_json(*args, **kwargs)

    def send_pyobj(self, *args, **kwargs):
        self.handle.send_pyobj(*args, **kwargs)

    def send_multipart(self, *args, **kwargs):
        self.handle.send_multipart(*args, **kwargs)

    def send_dumped(self, obj, *args, **kwargs):
        serialization = kwargs.pop(
            'serialization',
            self.conf.oslo_messaging_zmq.rpc_zmq_serialization)
        serializer = self._get_serializer(serialization)
        s = serializer.dump_as_bytes(obj)
        self.handle.send(s, *args, **kwargs)

    def recv(self, *args, **kwargs):
        return self.handle.recv(*args, **kwargs)

    def recv_string(self, *args, **kwargs):
        # NOTE(ozamiatin): Not using recv_string until
        # eventlet zmq support this convenience method
        # in thread-safe manner
        encoding = kwargs.pop('encoding', 'utf-8')
        s = self.handle.recv(*args, **kwargs)
        u = s.decode(encoding) if isinstance(s, six.binary_type) else s
        return u

    def recv_json(self, *args, **kwargs):
        return self.handle.recv_json(*args, **kwargs)

    def recv_pyobj(self, *args, **kwargs):
        return self.handle.recv_pyobj(*args, **kwargs)

    def recv_multipart(self, *args, **kwargs):
        return self.handle.recv_multipart(*args, **kwargs)

    def recv_loaded(self, *args, **kwargs):
        serialization = kwargs.pop(
            'serialization',
            self.conf.oslo_messaging_zmq.rpc_zmq_serialization)
        serializer = self._get_serializer(serialization)
        s = self.handle.recv(*args, **kwargs)
        obj = serializer.load_from_bytes(s)
        return obj

    def close(self, *args, **kwargs):
        identity = self.handle.identity
        self.handle.close(*args, **kwargs)
        LOG.debug("Socket %s closed" % identity)

    def connect_to_address(self, address):
        if address in self.connections:
            return
        stype = zmq_names.socket_type_str(self.socket_type)
        sid = self.handle.identity
        try:
            LOG.debug("Connecting %(stype)s socket %(sid)s to %(address)s",
                      {"stype": stype, "sid": sid, "address": address})
            self.connect(address)
        except zmq.ZMQError as e:
            LOG.error(_LE("Failed connecting %(stype)s-%(sid)s to "
                          "%(address)s: %(e)s"),
                      {"stype": stype, "sid": sid, "address": address, "e": e})
            raise rpc_common.RPCException(
                "Failed connecting %(stype)s-%(sid)s to %(address)s: %(e)s" %
                {"stype": stype, "sid": sid, "address": address, "e": e})

    def connect_to_host(self, host):
        address = zmq_address.get_tcp_direct_address(
            host.decode('utf-8') if six.PY3 and
            isinstance(host, six.binary_type) else host
        )
        self.connect_to_address(address)


class ZmqPortBusy(exceptions.MessagingException):
    """Raised when binding to a port failure"""

    def __init__(self, port_number):
        super(ZmqPortBusy, self).__init__()
        self.port_number = port_number


class ZmqRandomPortSocket(ZmqSocket):

    def __init__(self, conf, context, socket_type, host=None,
                 high_watermark=0, identity=None):
        super(ZmqRandomPortSocket, self).__init__(
            conf, context, socket_type, immediate=False,
            high_watermark=high_watermark, identity=identity)
        self.bind_address = zmq_address.get_tcp_random_address(self.conf)
        if host is None:
            host = conf.oslo_messaging_zmq.rpc_zmq_host
        try:
            self.port = self.handle.bind_to_random_port(
                self.bind_address,
                min_port=conf.oslo_messaging_zmq.rpc_zmq_min_port,
                max_port=conf.oslo_messaging_zmq.rpc_zmq_max_port,
                max_tries=conf.oslo_messaging_zmq.rpc_zmq_bind_port_retries)
            self.connect_address = zmq_address.combine_address(host, self.port)
        except zmq.ZMQBindError:
            LOG.error(_LE("Random ports range exceeded!"))
            raise ZmqPortBusy(port_number=0)


class ZmqFixedPortSocket(ZmqSocket):

    def __init__(self, conf, context, socket_type, host, port,
                 high_watermark=0, identity=None):
        super(ZmqFixedPortSocket, self).__init__(
            conf, context, socket_type, immediate=False,
            high_watermark=high_watermark, identity=identity)
        self.connect_address = zmq_address.combine_address(host, port)
        self.bind_address = zmq_address.get_tcp_direct_address(
            zmq_address.combine_address(
                conf.oslo_messaging_zmq.rpc_zmq_bind_address, port))
        self.host = host
        self.port = port

        try:
            self.handle.bind(self.bind_address)
        except zmq.ZMQError as e:
            LOG.exception(e)
            LOG.error(_LE("Chosen port %d is being busy.") % self.port)
            raise ZmqPortBusy(port_number=port)



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import uuid

import six

from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._drivers.zmq_driver import zmq_socket
from oslo_messaging._i18n import _LI, _LE

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


def check_message_format(func):
    def _check_message_format(socket):
        try:
            return func(socket)
        except Exception as e:
            LOG.error(_LE("Received message with wrong format: %r. "
                          "Dropping invalid message"), e)
            # NOTE(gdavoian): drop the left parts of a broken message, since
            # they most likely will break the order of next messages' parts
            if socket.getsockopt(zmq.RCVMORE):
                socket.recv_multipart()
    return _check_message_format


def create_socket(conf, context, port, socket_type):
    host = conf.zmq_proxy_opts.host
    identity = six.b(host) + b"/zmq-proxy/" + six.b(str(uuid.uuid4()))
    if port != 0:
        return zmq_socket.ZmqFixedPortSocket(conf, context, socket_type,
                                             host, port, identity=identity)
    else:
        return zmq_socket.ZmqRandomPortSocket(conf, context, socket_type,
                                              host, identity=identity)


class ProxyBase(object):

    PROXY_TYPE = "UNDEFINED"

    def __init__(self, conf, context, matchmaker):
        self.conf = conf
        self.context = context
        self.matchmaker = matchmaker

        LOG.info(_LI("Running %s proxy") % self.PROXY_TYPE)

        self.poller = zmq_async.get_poller()

    @staticmethod
    @check_message_format
    def _receive_message(socket):
        message = socket.recv_multipart()
        assert message[zmq_names.EMPTY_IDX] == b'', "Empty delimiter expected!"
        message_type = int(message[zmq_names.MESSAGE_TYPE_IDX])
        assert message_type in zmq_names.MESSAGE_TYPES, \
            "Known message type expected!"
        assert len(message) > zmq_names.MESSAGE_ID_IDX, \
            "At least %d parts expected!" % (zmq_names.MESSAGE_ID_IDX + 1)
        return message

    def cleanup(self):
        self.poller.close()



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import threading
import time

from oslo_messaging._drivers.zmq_driver import zmq_async

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


class TTLCache(object):

    _UNDEFINED = object()

    def __init__(self, ttl=None):
        self._lock = threading.Lock()
        self._cache = {}
        self._executor = None

        if not (ttl is None or isinstance(ttl, (int, float))):
            raise ValueError('ttl must be None or a number')

        # no (i.e. infinite) ttl
        if ttl is None or ttl <= 0:
            ttl = float('inf')
        else:
            self._executor = zmq_async.get_executor(self._update_cache)

        self._ttl = ttl

        if self._executor:
            self._executor.execute()

    @staticmethod
    def _is_expired(expiration_time, current_time):
        return expiration_time <= current_time

    def add(self, key, value=None):
        with self._lock:
            expiration_time = time.time() + self._ttl
            self._cache[key] = (value, expiration_time)

    def get(self, key, default=None):
        with self._lock:
            data = self._cache.get(key)
            if data is None:
                return default
            value, expiration_time = data
            if self._is_expired(expiration_time, time.time()):
                del self._cache[key]
                return default
            return value

    def __contains__(self, key):
        return self.get(key, self._UNDEFINED) is not self._UNDEFINED

    def _update_cache(self):
        with self._lock:
            current_time = time.time()
            old_size = len(self._cache)
            self._cache = \
                {key: (value, expiration_time) for
                 key, (value, expiration_time) in self._cache.items()
                 if not self._is_expired(expiration_time, current_time)}
            new_size = len(self._cache)
            LOG.debug('Updated cache: current size %(new_size)s '
                      '(%(size_difference)s records removed)',
                      {'new_size': new_size,
                       'size_difference': old_size - new_size})
        time.sleep(self._ttl)

    def cleanup(self):
        if self._executor:
            self._executor.stop()



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testscenarios
import testtools

import oslo_messaging
from oslo_messaging._drivers.zmq_driver import zmq_address
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging.tests import utils as test_utils


zmq = zmq_async.import_zmq()

load_tests = testscenarios.load_tests_apply_scenarios


class TestZmqAddress(test_utils.BaseTestCase):

    scenarios = [
        ('router', {'listener_type': zmq_names.socket_type_str(zmq.ROUTER)}),
        ('dealer', {'listener_type': zmq_names.socket_type_str(zmq.DEALER)})
    ]

    @testtools.skipIf(zmq is None, "zmq not available")
    def test_target_to_key_topic_only(self):
        target = oslo_messaging.Target(topic='topic')
        key = zmq_address.target_to_key(target, self.listener_type)
        self.assertEqual(self.listener_type + '/topic', key)

    @testtools.skipIf(zmq is None, "zmq not available")
    def test_target_to_key_topic_server_round_robin(self):
        target = oslo_messaging.Target(topic='topic', server='server')
        key = zmq_address.target_to_key(target, self.listener_type)
        self.assertEqual(self.listener_type + '/topic/server', key)

    @testtools.skipIf(zmq is None, "zmq not available")
    def test_target_to_key_topic_fanout(self):
        target = oslo_messaging.Target(topic='topic', fanout=True)
        key = zmq_address.target_to_key(target, self.listener_type)
        self.assertEqual(self.listener_type + '/topic', key)

    @testtools.skipIf(zmq is None, "zmq not available")
    def test_target_to_key_topic_server_fanout(self):
        target = oslo_messaging.Target(topic='topic', server='server',
                                       fanout=True)
        key = zmq_address.target_to_key(target, self.listener_type)
        self.assertEqual(self.listener_type + '/topic', key)

    @testtools.skipIf(zmq is None, "zmq not available")
    def test_target_to_key_topic_server_fanout_no_prefix(self):
        target = oslo_messaging.Target(topic='topic', server='server',
                                       fanout=True)
        key = zmq_address.target_to_key(target)
        self.assertEqual('topic', key)



#    Copyright 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import unittest

from concurrent import futures
from oslo_serialization import jsonutils
from oslo_utils import timeutils
import pika
from six.moves import mock

import oslo_messaging
from oslo_messaging._drivers.pika_driver import pika_commons as pika_drv_cmns
from oslo_messaging._drivers.pika_driver import pika_message as pika_drv_msg


class PikaIncomingMessageTestCase(unittest.TestCase):
    def setUp(self):
        self._pika_engine = mock.Mock()
        self._channel = mock.Mock()

        self._delivery_tag = 12345

        self._method = pika.spec.Basic.Deliver(delivery_tag=self._delivery_tag)
        self._properties = pika.BasicProperties(
            content_type="application/json",
            headers={"version": "1.0"},
        )
        self._body = (
            b'{"_$_key_context":"context_value",'
            b'"payload_key": "payload_value"}'
        )

    def test_message_body_parsing(self):
        message = pika_drv_msg.PikaIncomingMessage(
            self._pika_engine, self._channel, self._method, self._properties,
            self._body
        )

        self.assertEqual("context_value",
                         message.ctxt.get("key_context", None))
        self.assertEqual("payload_value",
                         message.message.get("payload_key", None))

    def test_message_acknowledge(self):
        message = pika_drv_msg.PikaIncomingMessage(
            self._pika_engine, self._channel, self._method, self._properties,
            self._body
        )

        message.acknowledge()

        self.assertEqual(1, self._channel.basic_ack.call_count)
        self.assertEqual({"delivery_tag": self._delivery_tag},
                         self._channel.basic_ack.call_args[1])

    def test_message_acknowledge_no_ack(self):
        message = pika_drv_msg.PikaIncomingMessage(
            self._pika_engine, None, self._method, self._properties,
            self._body
        )

        message.acknowledge()

        self.assertEqual(0, self._channel.basic_ack.call_count)

    def test_message_requeue(self):
        message = pika_drv_msg.PikaIncomingMessage(
            self._pika_engine, self._channel, self._method, self._properties,
            self._body
        )

        message.requeue()

        self.assertEqual(1, self._channel.basic_nack.call_count)
        self.assertEqual({"delivery_tag": self._delivery_tag, 'requeue': True},
                         self._channel.basic_nack.call_args[1])

    def test_message_requeue_no_ack(self):
        message = pika_drv_msg.PikaIncomingMessage(
            self._pika_engine, None, self._method, self._properties,
            self._body
        )

        message.requeue()

        self.assertEqual(0, self._channel.basic_nack.call_count)


class RpcPikaIncomingMessageTestCase(unittest.TestCase):
    def setUp(self):
        self._pika_engine = mock.Mock()
        self._pika_engine.rpc_reply_retry_attempts = 3
        self._pika_engine.rpc_reply_retry_delay = 0.25

        self._channel = mock.Mock()

        self._delivery_tag = 12345

        self._method = pika.spec.Basic.Deliver(delivery_tag=self._delivery_tag)
        self._body = (
            b'{"_$_key_context":"context_value",'
            b'"payload_key":"payload_value"}'
        )
        self._properties = pika.BasicProperties(
            content_type="application/json",
            headers={"version": "1.0"},
        )

    def test_call_message_body_parsing(self):
        self._properties.correlation_id = 123456789
        self._properties.reply_to = "reply_queue"

        message = pika_drv_msg.RpcPikaIncomingMessage(
            self._pika_engine, self._channel, self._method, self._properties,
            self._body
        )

        self.assertEqual("context_value",
                         message.ctxt.get("key_context", None))
        self.assertEqual(123456789, message.msg_id)
        self.assertEqual("reply_queue", message.reply_q)

        self.assertEqual("payload_value",
                         message.message.get("payload_key", None))

    def test_cast_message_body_parsing(self):
        message = pika_drv_msg.RpcPikaIncomingMessage(
            self._pika_engine, self._channel, self._method, self._properties,
            self._body
        )

        self.assertEqual("context_value",
                         message.ctxt.get("key_context", None))
        self.assertIsNone(message.msg_id)
        self.assertIsNone(message.reply_q)

        self.assertEqual("payload_value",
                         message.message.get("payload_key", None))

    @mock.patch(("oslo_messaging._drivers.pika_driver.pika_message."
                 "PikaOutgoingMessage.send"))
    def test_reply_for_cast_message(self, send_reply_mock):
        message = pika_drv_msg.RpcPikaIncomingMessage(
            self._pika_engine, self._channel, self._method, self._properties,
            self._body
        )

        self.assertEqual("context_value",
                         message.ctxt.get("key_context", None))
        self.assertIsNone(message.msg_id)
        self.assertIsNone(message.reply_q)

        self.assertEqual("payload_value",
                         message.message.get("payload_key", None))

        message.reply(reply=object())

        self.assertEqual(0, send_reply_mock.call_count)

    @mock.patch("oslo_messaging._drivers.pika_driver.pika_message."
                "RpcReplyPikaOutgoingMessage")
    @mock.patch("tenacity.retry")
    def test_positive_reply_for_call_message(self,
                                             retry_mock,
                                             outgoing_message_mock):
        self._properties.correlation_id = 123456789
        self._properties.reply_to = "reply_queue"

        message = pika_drv_msg.RpcPikaIncomingMessage(
            self._pika_engine, self._channel, self._method, self._properties,
            self._body
        )

        self.assertEqual("context_value",
                         message.ctxt.get("key_context", None))
        self.assertEqual(123456789, message.msg_id)
        self.assertEqual("reply_queue", message.reply_q)

        self.assertEqual("payload_value",
                         message.message.get("payload_key", None))
        reply = "all_fine"
        message.reply(reply=reply)

        outgoing_message_mock.assert_called_once_with(
            self._pika_engine, 123456789, failure_info=None, reply='all_fine',
            content_type='application/json'
        )
        outgoing_message_mock().send.assert_called_once_with(
            reply_q='reply_queue', stopwatch=mock.ANY, retrier=mock.ANY
        )
        retry_mock.assert_called_once_with(
            stop=mock.ANY, retry=mock.ANY, wait=mock.ANY
        )

    @mock.patch("oslo_messaging._drivers.pika_driver.pika_message."
                "RpcReplyPikaOutgoingMessage")
    @mock.patch("tenacity.retry")
    def test_negative_reply_for_call_message(self,
                                             retry_mock,
                                             outgoing_message_mock):
        self._properties.correlation_id = 123456789
        self._properties.reply_to = "reply_queue"

        message = pika_drv_msg.RpcPikaIncomingMessage(
            self._pika_engine, self._channel, self._method, self._properties,
            self._body
        )

        self.assertEqual("context_value",
                         message.ctxt.get("key_context", None))
        self.assertEqual(123456789, message.msg_id)
        self.assertEqual("reply_queue", message.reply_q)

        self.assertEqual("payload_value",
                         message.message.get("payload_key", None))

        failure_info = object()
        message.reply(failure=failure_info)

        outgoing_message_mock.assert_called_once_with(
            self._pika_engine, 123456789,
            failure_info=failure_info,
            reply=None,
            content_type='application/json'
        )
        outgoing_message_mock().send.assert_called_once_with(
            reply_q='reply_queue', stopwatch=mock.ANY, retrier=mock.ANY
        )
        retry_mock.assert_called_once_with(
            stop=mock.ANY, retry=mock.ANY, wait=mock.ANY
        )


class RpcReplyPikaIncomingMessageTestCase(unittest.TestCase):
    def setUp(self):
        self._pika_engine = mock.Mock()
        self._pika_engine.allowed_remote_exmods = [
            pika_drv_cmns.EXCEPTIONS_MODULE, "oslo_messaging.exceptions"
        ]

        self._channel = mock.Mock()

        self._delivery_tag = 12345

        self._method = pika.spec.Basic.Deliver(delivery_tag=self._delivery_tag)

        self._properties = pika.BasicProperties(
            content_type="application/json",
            headers={"version": "1.0"},
            correlation_id=123456789
        )

    def test_positive_reply_message_body_parsing(self):

        body = b'{"s": "all fine"}'

        message = pika_drv_msg.RpcReplyPikaIncomingMessage(
            self._pika_engine, self._channel, self._method, self._properties,
            body
        )

        self.assertEqual(123456789, message.msg_id)
        self.assertIsNone(message.failure)
        self.assertEqual("all fine", message.result)

    def test_negative_reply_message_body_parsing(self):

        body = (b'{'
                b'    "e": {'
                b'         "s": "Error message",'
                b'         "t": ["TRACE HERE"],'
                b'         "c": "MessagingException",'
                b'         "m": "oslo_messaging.exceptions"'
                b'     }'
                b'}')

        message = pika_drv_msg.RpcReplyPikaIncomingMessage(
            self._pika_engine, self._channel, self._method, self._properties,
            body
        )

        self.assertEqual(123456789, message.msg_id)
        self.assertIsNone(message.result)
        self.assertEqual(
            'Error message\n'
            'TRACE HERE',
            str(message.failure)
        )
        self.assertIsInstance(message.failure,
                              oslo_messaging.MessagingException)


class PikaOutgoingMessageTestCase(unittest.TestCase):
    def setUp(self):
        self._pika_engine = mock.MagicMock()
        self._pika_engine.default_content_type = "application/json"
        self._exchange = "it is exchange"
        self._routing_key = "it is routing key"
        self._expiration = 1
        self._stopwatch = (
            timeutils.StopWatch(duration=self._expiration).start()
        )
        self._mandatory = object()

        self._message = {"msg_type": 1, "msg_str": "hello"}
        self._context = {"request_id": 555, "token": "it is a token"}

    @mock.patch("oslo_serialization.jsonutils.dump_as_bytes",
                new=functools.partial(jsonutils.dump_as_bytes, sort_keys=True))
    def test_send_with_confirmation(self):
        message = pika_drv_msg.PikaOutgoingMessage(
            self._pika_engine, self._message, self._context
        )

        message.send(
            exchange=self._exchange,
            routing_key=self._routing_key,
            confirm=True,
            mandatory=self._mandatory,
            persistent=True,
            stopwatch=self._stopwatch,
            retrier=None
        )

        self._pika_engine.connection_with_confirmation_pool.acquire(
        ).__enter__().channel.publish.assert_called_once_with(
            body=mock.ANY,
            exchange=self._exchange, mandatory=self._mandatory,
            properties=mock.ANY,
            routing_key=self._routing_key
        )

        body = self._pika_engine.connection_with_confirmation_pool.acquire(
        ).__enter__().channel.publish.call_args[1]["body"]

        self.assertEqual(
            b'{"_$_request_id": 555, "_$_token": "it is a token", '
            b'"msg_str": "hello", "msg_type": 1}',
            body
        )

        props = self._pika_engine.connection_with_confirmation_pool.acquire(
        ).__enter__().channel.publish.call_args[1]["properties"]

        self.assertEqual('application/json', props.content_type)
        self.assertEqual(2, props.delivery_mode)
        self.assertTrue(self._expiration * 1000 - float(props.expiration) <
                        100)
        self.assertEqual({'version': '1.0'}, props.headers)
        self.assertTrue(props.message_id)

    @mock.patch("oslo_serialization.jsonutils.dump_as_bytes",
                new=functools.partial(jsonutils.dump_as_bytes, sort_keys=True))
    def test_send_without_confirmation(self):
        message = pika_drv_msg.PikaOutgoingMessage(
            self._pika_engine, self._message, self._context
        )

        message.send(
            exchange=self._exchange,
            routing_key=self._routing_key,
            confirm=False,
            mandatory=self._mandatory,
            persistent=False,
            stopwatch=self._stopwatch,
            retrier=None
        )

        self._pika_engine.connection_without_confirmation_pool.acquire(
        ).__enter__().channel.publish.assert_called_once_with(
            body=mock.ANY,
            exchange=self._exchange, mandatory=self._mandatory,
            properties=mock.ANY,
            routing_key=self._routing_key
        )

        body = self._pika_engine.connection_without_confirmation_pool.acquire(
        ).__enter__().channel.publish.call_args[1]["body"]

        self.assertEqual(
            b'{"_$_request_id": 555, "_$_token": "it is a token", '
            b'"msg_str": "hello", "msg_type": 1}',
            body
        )

        props = self._pika_engine.connection_without_confirmation_pool.acquire(
        ).__enter__().channel.publish.call_args[1]["properties"]

        self.assertEqual('application/json', props.content_type)
        self.assertEqual(1, props.delivery_mode)
        self.assertTrue(self._expiration * 1000 - float(props.expiration)
                        < 100)
        self.assertEqual({'version': '1.0'}, props.headers)
        self.assertTrue(props.message_id)


class RpcPikaOutgoingMessageTestCase(unittest.TestCase):
    def setUp(self):
        self._exchange = "it is exchange"
        self._routing_key = "it is routing key"

        self._pika_engine = mock.MagicMock()
        self._pika_engine.get_rpc_exchange_name.return_value = self._exchange
        self._pika_engine.get_rpc_queue_name.return_value = self._routing_key
        self._pika_engine.default_content_type = "application/json"

        self._message = {"msg_type": 1, "msg_str": "hello"}
        self._context = {"request_id": 555, "token": "it is a token"}

    @mock.patch("oslo_serialization.jsonutils.dump_as_bytes",
                new=functools.partial(jsonutils.dump_as_bytes, sort_keys=True))
    def test_send_cast_message(self):
        message = pika_drv_msg.RpcPikaOutgoingMessage(
            self._pika_engine, self._message, self._context
        )

        expiration = 1
        stopwatch = timeutils.StopWatch(duration=expiration).start()

        message.send(
            exchange=self._exchange,
            routing_key=self._routing_key,
            reply_listener=None,
            stopwatch=stopwatch,
            retrier=None
        )

        self._pika_engine.connection_with_confirmation_pool.acquire(
        ).__enter__().channel.publish.assert_called_once_with(
            body=mock.ANY,
            exchange=self._exchange, mandatory=True,
            properties=mock.ANY,
            routing_key=self._routing_key
        )

        body = self._pika_engine.connection_with_confirmation_pool.acquire(
        ).__enter__().channel.publish.call_args[1]["body"]

        self.assertEqual(
            b'{"_$_request_id": 555, "_$_token": "it is a token", '
            b'"msg_str": "hello", "msg_type": 1}',
            body
        )

        props = self._pika_engine.connection_with_confirmation_pool.acquire(
        ).__enter__().channel.publish.call_args[1]["properties"]

        self.assertEqual('application/json', props.content_type)
        self.assertEqual(1, props.delivery_mode)
        self.assertTrue(expiration * 1000 - float(props.expiration) < 100)
        self.assertEqual({'version': '1.0'}, props.headers)
        self.assertIsNone(props.correlation_id)
        self.assertIsNone(props.reply_to)
        self.assertTrue(props.message_id)

    @mock.patch("oslo_serialization.jsonutils.dump_as_bytes",
                new=functools.partial(jsonutils.dump_as_bytes, sort_keys=True))
    def test_send_call_message(self):
        message = pika_drv_msg.RpcPikaOutgoingMessage(
            self._pika_engine, self._message, self._context
        )

        expiration = 1
        stopwatch = timeutils.StopWatch(duration=expiration).start()

        result = "it is a result"
        reply_queue_name = "reply_queue_name"

        future = futures.Future()
        future.set_result(result)
        reply_listener = mock.Mock()
        reply_listener.register_reply_waiter.return_value = future
        reply_listener.get_reply_qname.return_value = reply_queue_name

        res = message.send(
            exchange=self._exchange,
            routing_key=self._routing_key,
            reply_listener=reply_listener,
            stopwatch=stopwatch,
            retrier=None
        )

        self.assertEqual(result, res)

        self._pika_engine.connection_with_confirmation_pool.acquire(
        ).__enter__().channel.publish.assert_called_once_with(
            body=mock.ANY,
            exchange=self._exchange, mandatory=True,
            properties=mock.ANY,
            routing_key=self._routing_key
        )

        body = self._pika_engine.connection_with_confirmation_pool.acquire(
        ).__enter__().channel.publish.call_args[1]["body"]

        self.assertEqual(
            b'{"_$_request_id": 555, "_$_token": "it is a token", '
            b'"msg_str": "hello", "msg_type": 1}',
            body
        )

        props = self._pika_engine.connection_with_confirmation_pool.acquire(
        ).__enter__().channel.publish.call_args[1]["properties"]

        self.assertEqual('application/json', props.content_type)
        self.assertEqual(1, props.delivery_mode)
        self.assertTrue(expiration * 1000 - float(props.expiration) < 100)
        self.assertEqual({'version': '1.0'}, props.headers)
        self.assertEqual(message.msg_id, props.correlation_id)
        self.assertEqual(reply_queue_name, props.reply_to)
        self.assertTrue(props.message_id)


class RpcReplyPikaOutgoingMessageTestCase(unittest.TestCase):
    def setUp(self):
        self._reply_q = "reply_queue_name"

        self._expiration = 1
        self._stopwatch = (
            timeutils.StopWatch(duration=self._expiration).start()
        )

        self._pika_engine = mock.MagicMock()

        self._rpc_reply_exchange = "rpc_reply_exchange"
        self._pika_engine.rpc_reply_exchange = self._rpc_reply_exchange
        self._pika_engine.default_content_type = "application/json"

        self._msg_id = 12345567

    @mock.patch("oslo_serialization.jsonutils.dump_as_bytes",
                new=functools.partial(jsonutils.dump_as_bytes, sort_keys=True))
    def test_success_message_send(self):
        message = pika_drv_msg.RpcReplyPikaOutgoingMessage(
            self._pika_engine, self._msg_id, reply="all_fine"
        )

        message.send(self._reply_q, stopwatch=self._stopwatch, retrier=None)

        self._pika_engine.connection_with_confirmation_pool.acquire(
        ).__enter__().channel.publish.assert_called_once_with(
            body=b'{"s": "all_fine"}',
            exchange=self._rpc_reply_exchange, mandatory=True,
            properties=mock.ANY,
            routing_key=self._reply_q
        )

        props = self._pika_engine.connection_with_confirmation_pool.acquire(
        ).__enter__().channel.publish.call_args[1]["properties"]

        self.assertEqual('application/json', props.content_type)
        self.assertEqual(1, props.delivery_mode)
        self.assertTrue(self._expiration * 1000 - float(props.expiration) <
                        100)
        self.assertEqual({'version': '1.0'}, props.headers)
        self.assertEqual(message.msg_id, props.correlation_id)
        self.assertIsNone(props.reply_to)
        self.assertTrue(props.message_id)

    @mock.patch("traceback.format_exception", new=lambda x, y, z: z)
    @mock.patch("oslo_serialization.jsonutils.dump_as_bytes",
                new=functools.partial(jsonutils.dump_as_bytes, sort_keys=True))
    def test_failure_message_send(self):
        failure_info = (oslo_messaging.MessagingException,
                        oslo_messaging.MessagingException("Error message"),
                        ['It is a trace'])

        message = pika_drv_msg.RpcReplyPikaOutgoingMessage(
            self._pika_engine, self._msg_id, failure_info=failure_info
        )

        message.send(self._reply_q, stopwatch=self._stopwatch, retrier=None)

        self._pika_engine.connection_with_confirmation_pool.acquire(
        ).__enter__().channel.publish.assert_called_once_with(
            body=mock.ANY,
            exchange=self._rpc_reply_exchange,
            mandatory=True,
            properties=mock.ANY,
            routing_key=self._reply_q
        )

        body = self._pika_engine.connection_with_confirmation_pool.acquire(
        ).__enter__().channel.publish.call_args[1]["body"]
        self.assertEqual(
            b'{"e": {"c": "MessagingException", '
            b'"m": "oslo_messaging.exceptions", "s": "Error message", '
            b'"t": ["It is a trace"]}}',
            body
        )

        props = self._pika_engine.connection_with_confirmation_pool.acquire(
        ).__enter__().channel.publish.call_args[1]["properties"]

        self.assertEqual('application/json', props.content_type)
        self.assertEqual(1, props.delivery_mode)
        self.assertTrue(self._expiration * 1000 - float(props.expiration) <
                        100)
        self.assertEqual({'version': '1.0'}, props.headers)
        self.assertEqual(message.msg_id, props.correlation_id)
        self.assertIsNone(props.reply_to)
        self.assertTrue(props.message_id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from six.moves import mock
import testtools

from oslo_messaging._drivers.zmq_driver.poller import green_poller
from oslo_messaging._drivers.zmq_driver.poller import threading_poller
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging.tests import utils as test_utils

zmq = zmq_async.import_zmq()


class TestImportZmq(test_utils.BaseTestCase):

    @testtools.skipIf(zmq is None, "zmq not available")
    def setUp(self):
        super(TestImportZmq, self).setUp()

    def test_when_eventlet_is_available_then_load_eventlet_green_zmq(self):
        zmq_async.eventletutils.is_monkey_patched = lambda _: True

        mock_try_import = mock.Mock()
        zmq_async.importutils.try_import = mock_try_import

        zmq_async.import_zmq()

        mock_try_import.assert_called_with('eventlet.green.zmq', default=None)

    def test_when_evetlet_is_unavailable_then_load_zmq(self):
        zmq_async.eventletutils.is_monkey_patched = lambda _: False

        mock_try_import = mock.Mock()
        zmq_async.importutils.try_import = mock_try_import

        zmq_async.import_zmq()

        mock_try_import.assert_called_with('zmq', default=None)


class TestGetPoller(test_utils.BaseTestCase):

    @testtools.skipIf(zmq is None, "zmq not available")
    def setUp(self):
        super(TestGetPoller, self).setUp()

    def test_when_eventlet_is_available_then_return_GreenPoller(self):
        zmq_async.eventletutils.is_monkey_patched = lambda _: True

        poller = zmq_async.get_poller()

        self.assertIsInstance(poller, green_poller.GreenPoller)

    def test_when_eventlet_is_unavailable_then_return_ThreadingPoller(self):
        zmq_async.eventletutils.is_monkey_patched = lambda _: False

        poller = zmq_async.get_poller()

        self.assertIsInstance(poller, threading_poller.ThreadingPoller)


class TestGetExecutor(test_utils.BaseTestCase):

    @testtools.skipIf(zmq is None, "zmq not available")
    def setUp(self):
        super(TestGetExecutor, self).setUp()

    def test_when_eventlet_module_is_available_then_return_GreenExecutor(self):
        zmq_async.eventletutils.is_monkey_patched = lambda _: True

        executor = zmq_async.get_executor('any method')

        self.assertIsInstance(executor, green_poller.GreenExecutor)
        self.assertEqual('any method', executor._method)

    def test_when_eventlet_is_unavailable_then_return_ThreadingExecutor(self):
        zmq_async.eventletutils.is_monkey_patched = lambda _: False

        executor = zmq_async.get_executor('any method')

        self.assertIsInstance(executor,
                              threading_poller.ThreadingExecutor)
        self.assertEqual('any method', executor._method)




# Copyright 2013 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testscenarios

import oslo_messaging
from oslo_messaging.tests import utils as test_utils

load_tests = testscenarios.load_tests_apply_scenarios


class TargetConstructorTestCase(test_utils.BaseTestCase):

    scenarios = [
        ('all_none', dict(kwargs=dict())),
        ('exchange', dict(kwargs=dict(exchange='testexchange'))),
        ('topic', dict(kwargs=dict(topic='testtopic'))),
        ('namespace', dict(kwargs=dict(namespace='testnamespace'))),
        ('version', dict(kwargs=dict(version='3.4'))),
        ('server', dict(kwargs=dict(server='testserver'))),
        ('fanout', dict(kwargs=dict(fanout=True))),
    ]

    def test_constructor(self):
        target = oslo_messaging.Target(**self.kwargs)
        for k in self.kwargs:
            self.assertEqual(self.kwargs[k], getattr(target, k))
        for k in ['exchange', 'topic', 'namespace',
                  'version', 'server', 'fanout']:
            if k in self.kwargs:
                continue
            self.assertIsNone(getattr(target, k))


class TargetCallableTestCase(test_utils.BaseTestCase):

    scenarios = [
        ('all_none', dict(attrs=dict(), kwargs=dict(), vals=dict())),
        ('exchange_attr', dict(attrs=dict(exchange='testexchange'),
                               kwargs=dict(),
                               vals=dict(exchange='testexchange'))),
        ('exchange_arg', dict(attrs=dict(),
                              kwargs=dict(exchange='testexchange'),
                              vals=dict(exchange='testexchange'))),
        ('topic_attr', dict(attrs=dict(topic='testtopic'),
                            kwargs=dict(),
                            vals=dict(topic='testtopic'))),
        ('topic_arg', dict(attrs=dict(),
                           kwargs=dict(topic='testtopic'),
                           vals=dict(topic='testtopic'))),
        ('namespace_attr', dict(attrs=dict(namespace='testnamespace'),
                                kwargs=dict(),
                                vals=dict(namespace='testnamespace'))),
        ('namespace_arg', dict(attrs=dict(),
                               kwargs=dict(namespace='testnamespace'),
                               vals=dict(namespace='testnamespace'))),
        ('version_attr', dict(attrs=dict(version='3.4'),
                              kwargs=dict(),
                              vals=dict(version='3.4'))),
        ('version_arg', dict(attrs=dict(),
                             kwargs=dict(version='3.4'),
                             vals=dict(version='3.4'))),
        ('server_attr', dict(attrs=dict(server='testserver'),
                             kwargs=dict(),
                             vals=dict(server='testserver'))),
        ('server_arg', dict(attrs=dict(),
                            kwargs=dict(server='testserver'),
                            vals=dict(server='testserver'))),
        ('fanout_attr', dict(attrs=dict(fanout=True),
                             kwargs=dict(),
                             vals=dict(fanout=True))),
        ('fanout_arg', dict(attrs=dict(),
                            kwargs=dict(fanout=True),
                            vals=dict(fanout=True))),
    ]

    def test_callable(self):
        target = oslo_messaging.Target(**self.attrs)
        target = target(**self.kwargs)
        for k in self.vals:
            self.assertEqual(self.vals[k], getattr(target, k))
        for k in ['exchange', 'topic', 'namespace',
                  'version', 'server', 'fanout']:
            if k in self.vals:
                continue
            self.assertIsNone(getattr(target, k))


class TargetReprTestCase(test_utils.BaseTestCase):

    scenarios = [
        ('all_none', dict(kwargs=dict(), repr='')),
        ('exchange', dict(kwargs=dict(exchange='testexchange'),
                          repr='exchange=testexchange')),
        ('topic', dict(kwargs=dict(topic='testtopic'),
                       repr='topic=testtopic')),
        ('namespace', dict(kwargs=dict(namespace='testnamespace'),
                           repr='namespace=testnamespace')),
        ('version', dict(kwargs=dict(version='3.4'),
                         repr='version=3.4')),
        ('server', dict(kwargs=dict(server='testserver'),
                        repr='server=testserver')),
        ('fanout', dict(kwargs=dict(fanout=True),
                        repr='fanout=True')),
        ('exchange_and_fanout', dict(kwargs=dict(exchange='testexchange',
                                                 fanout=True),
                                     repr='exchange=testexchange, '
                                          'fanout=True')),
    ]

    def test_repr(self):
        target = oslo_messaging.Target(**self.kwargs)
        self.assertEqual('<Target ' + self.repr + '>', str(target))


_notset = object()


class EqualityTestCase(test_utils.BaseTestCase):

    @classmethod
    def generate_scenarios(cls):
        attr = [
            ('exchange', dict(attr='exchange')),
            ('topic', dict(attr='topic')),
            ('namespace', dict(attr='namespace')),
            ('version', dict(attr='version')),
            ('server', dict(attr='server')),
            ('fanout', dict(attr='fanout')),
        ]
        a = [
            ('a_notset', dict(a_value=_notset)),
            ('a_none', dict(a_value=None)),
            ('a_empty', dict(a_value='')),
            ('a_foo', dict(a_value='foo')),
            ('a_bar', dict(a_value='bar')),
        ]
        b = [
            ('b_notset', dict(b_value=_notset)),
            ('b_none', dict(b_value=None)),
            ('b_empty', dict(b_value='')),
            ('b_foo', dict(b_value='foo')),
            ('b_bar', dict(b_value='bar')),
        ]

        cls.scenarios = testscenarios.multiply_scenarios(attr, a, b)
        for s in cls.scenarios:
            s[1]['equals'] = (s[1]['a_value'] == s[1]['b_value'])

    def test_equality(self):
        a_kwargs = {self.attr: self.a_value}
        b_kwargs = {self.attr: self.b_value}

        a = oslo_messaging.Target(**a_kwargs)
        b = oslo_messaging.Target(**b_kwargs)

        if self.equals:
            self.assertEqual(a, b)
            self.assertFalse(a != b)
        else:
            self.assertNotEqual(a, b)
            self.assertFalse(a == b)


EqualityTestCase.generate_scenarios()



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from oslo_messaging._drivers.zmq_driver.server import zmq_ttl_cache
from oslo_messaging.tests import utils as test_utils


class TestZmqTTLCache(test_utils.BaseTestCase):

    def setUp(self):
        super(TestZmqTTLCache, self).setUp()

        def call_count_decorator(unbound_method):
            def wrapper(self, *args, **kwargs):
                wrapper.call_count += 1
                return unbound_method(self, *args, **kwargs)
            wrapper.call_count = 0
            return wrapper

        zmq_ttl_cache.TTLCache._update_cache = \
            call_count_decorator(zmq_ttl_cache.TTLCache._update_cache)

        self.cache = zmq_ttl_cache.TTLCache(ttl=1)

        self.addCleanup(lambda: self.cache.cleanup())

    def _test_add_get(self):
        self.cache.add('x', 'a')

        self.assertEqual(self.cache.get('x'), 'a')
        self.assertEqual(self.cache.get('x', 'b'), 'a')
        self.assertIsNone(self.cache.get('y'))
        self.assertEqual(self.cache.get('y', 'b'), 'b')

        time.sleep(1)

        self.assertIsNone(self.cache.get('x'))
        self.assertEqual(self.cache.get('x', 'b'), 'b')

    def test_add_get_with_executor(self):
        self._test_add_get()

    def test_add_get_without_executor(self):
        self.cache._executor.stop()
        self._test_add_get()

    def _test_in_operator(self):
        self.cache.add(1)

        self.assertIn(1, self.cache)

        time.sleep(0.5)

        self.cache.add(2)

        self.assertIn(1, self.cache)
        self.assertIn(2, self.cache)

        time.sleep(0.75)

        self.cache.add(3)

        self.assertNotIn(1, self.cache)
        self.assertIn(2, self.cache)
        self.assertIn(3, self.cache)

        time.sleep(0.5)

        self.assertNotIn(2, self.cache)
        self.assertIn(3, self.cache)

    def test_in_operator_with_executor(self):
        self._test_in_operator()

    def test_in_operator_without_executor(self):
        self.cache._executor.stop()
        self._test_in_operator()

    def _is_expired(self, key):
        with self.cache._lock:
            _, expiration_time = self.cache._cache[key]
            return self.cache._is_expired(expiration_time, time.time())

    def test_executor(self):
        self.cache.add(1)

        self.assertEqual([1], sorted(self.cache._cache.keys()))
        self.assertFalse(self._is_expired(1))

        time.sleep(0.75)

        self.assertEqual(1, self.cache._update_cache.call_count)

        self.cache.add(2)

        self.assertEqual([1, 2], sorted(self.cache._cache.keys()))
        self.assertFalse(self._is_expired(1))
        self.assertFalse(self._is_expired(2))

        time.sleep(0.75)

        self.assertEqual(2, self.cache._update_cache.call_count)

        self.cache.add(3)

        if 1 in self.cache:
            self.assertEqual([1, 2, 3], sorted(self.cache._cache.keys()))
            self.assertTrue(self._is_expired(1))
        else:
            self.assertEqual([2, 3], sorted(self.cache._cache.keys()))
        self.assertFalse(self._is_expired(2))
        self.assertFalse(self._is_expired(3))

        time.sleep(0.75)

        self.assertEqual(3, self.cache._update_cache.call_count)

        self.assertEqual([3], sorted(self.cache._cache.keys()))
        self.assertFalse(self._is_expired(3))



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_messaging._drivers.zmq_driver import zmq_version
from oslo_messaging.tests import utils as test_utils


class Doer(object):

    def __init__(self):
        self.x = 1
        self.y = 2
        self.z = 3

    def _sudo(self):
        pass

    def do(self):
        pass

    def _do_v_1_1(self):
        pass

    def _do_v_2_2(self):
        pass

    def _do_v_3_3(self):
        pass


class TestZmqVersion(test_utils.BaseTestCase):

    def setUp(self):
        super(TestZmqVersion, self).setUp()
        self.doer = Doer()

    def test_get_unknown_attr_versions(self):
        self.assertRaises(AssertionError, zmq_version.get_method_versions,
                          self.doer, 'qwerty')

    def test_get_non_method_attr_versions(self):
        for attr_name in vars(self.doer):
            self.assertRaises(AssertionError, zmq_version.get_method_versions,
                              self.doer, attr_name)

    def test_get_private_method_versions(self):
        self.assertRaises(AssertionError, zmq_version.get_method_versions,
                          self.doer, '_sudo')

    def test_get_public_method_versions(self):
        do_versions = zmq_version.get_method_versions(self.doer, 'do')
        self.assertEqual(['1.1', '2.2', '3.3'], sorted(do_versions.keys()))
